@@ -8,7 +8,7 @@ use Carp;
 use vars qw( $SOCKET_CLASS @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION $SOCKS_ERROR $SOCKS5_RESOLVE $SOCKS4_RESOLVE $SOCKS_DEBUG %CODES );
 require Exporter;
 
-$VERSION = '0.68';
+$VERSION = '0.69';
 
 use constant {
 	SOCKS_WANT_READ  => 20,
@@ -325,11 +325,13 @@ sub _configure {
 		}
 		${*$self}->{SOCKS}->{CmdAddr} = delete($args->{UdpAddr});
 		${*$self}->{SOCKS}->{CmdPort} = delete($args->{UdpPort});
-		$args->{LocalAddr} = ${*$self}->{SOCKS}->{CmdAddr};
-		$args->{LocalPort} = ${*$self}->{SOCKS}->{CmdPort};
 		${*$self}->{SOCKS}->{TCP} = __PACKAGE__->new(    # TCP backend for UDP socket
-			Timeout => $args->{Timeout},
-			Proto   => 'tcp'
+			Timeout  => $args->{Timeout},
+			Proto    => 'tcp',
+			PeerAddr => $args->{ProxyAddr},
+			PeerPort => $args->{ProxyPort},
+			exists $args->{Blocking} ?
+				(Blocking => $args->{Blocking}) : ()
 		) or return;
 	}
 	elsif (exists($args->{ConnectAddr}) && exists($args->{ConnectPort})) {
@@ -353,10 +355,11 @@ sub connect {
 
 	my $ok =
 	  defined(${*$self}->{SOCKS}->{TCP})
-	  ? ${*$self}->{SOCKS}->{TCP}->SUPER::connect(@_)
+	  ? 1
 	  : $self->SUPER::connect(@_);
 
-	if (($! == EINPROGRESS || $! == EWOULDBLOCK) && $self->blocking == 0) {
+	if (($! == EINPROGRESS || $! == EWOULDBLOCK) && 
+	    (${*$self}->{SOCKS}->{TCP} || $self)->blocking == 0) {
 		$SOCKS_ERROR->set(SOCKS_WANT_WRITE, 'Socks want write');
 	}
 	elsif (!$ok) {
@@ -368,6 +371,9 @@ sub connect {
 		$SOCKS_ERROR->set();
 	}
 
+	return $ok # proxy address was not specified, so do not make socks handshake
+		unless ${*$self}->{SOCKS}->{ProxyAddr} && ${*$self}->{SOCKS}->{ProxyPort};
+	
 	$self->_connect();
 }
 
@@ -406,10 +412,33 @@ sub _connect {
 		defined($self->_run_queue())
 		  or return;
 	}
+	else {
+		if ($self->isa('IO::Socket::IP')) {
+			# IO::Socket::IP requires multiple connect calls
+			# when performing non-blocking multi-homed connect
+			unshift @{ ${*$self}->{SOCKS}->{queue} }, ['_socket_connect', [], undef, [], 0];
+		}
+		
+		return; # connect() return value
+	}
 
 	return $self;
 }
 
+sub _socket_connect {
+	my $self = shift;
+	my $sock = ${*$self}->{SOCKS}->{TCP} || $self;
+	
+	return 1 if $sock->SUPER::connect();
+	if ($! == EINPROGRESS || $! == EWOULDBLOCK) {
+		$SOCKS_ERROR->set(SOCKS_WANT_WRITE, 'Socks want write');
+		return -1;
+	}
+	
+	$SOCKS_ERROR->set($!, $@ = "Connection to proxy failed: $!");
+	return;
+}
+
 sub _run_queue {
 	# run tasks from queue, return undef on error, -1 if one of the task
     # returned not completed because of the possible blocking on network operation
@@ -428,7 +457,7 @@ sub _run_queue {
 		}
 
 		last if ($retval == -1);
-		${*$self}->{SOCKS}->{queue_results}{ $elt->[Q_SUB] } = $retval;
+		${*$self}->{SOCKS}->{queue_results}{$sub} = $retval;
 		if ($elt->[Q_OKCB]) {
 			$elt->[Q_OKCB]->();
 		}
@@ -2115,11 +2144,13 @@ Both takes the following config hash:
   BindPort => Port of the remote machine which will
               connect to the proxy server after bind request
   
-  UdpAddr => Associate UDP socket on the server with this client
-             hostname
+  UdpAddr => Expected address where datagrams will be sent. Fill it with address
+             of all zeros if address is not known at this moment.
+             Proxy server may use this information to limit access to the association.
   
-  UdpPort => Associate UDP socket on the server with this client
-             port
+  UdpPort => Expected port where datagrams will be sent. Use zero port
+             if port is not known at this moment. Proxy server may use this
+             information to limit access to the association.
   
   AuthType => What kind of authentication to support:
               none       - no authentication (default)
@@ -2209,6 +2240,12 @@ Example:
         else {
             die $SOCKS_ERROR;
         }
+        
+        # NOTE: when base class ($IO::Socket::Socks::SOCKET_CLASS) is IO::Socket::IP
+        # and you are using kqueue or epoll to check for readable/writable sockets
+        # you need to readd $sock to kqueue/epoll after each call to ready() (actually until socket will be connected to proxy server),
+        # because IO::Socket::IP may change internal socket of $sock for milti-homed hosts.
+        # There is no such problem when you are using select/poll
     }
     
     # you may want to return socket to blocking state by $sock->blocking(1)
